# -*- coding: utf-8 -*-
__author__ = "Benedict J. Heinen"
__copyright__ = "Copyright 2018, Benedict J. Heinen"
__email__ = "benedict.heinen@gmail.com"

import os
import numpy as np
from scipy.optimize import minimize
from PyQt5.QtCore import Qt, pyqtSignal
from PyQt5.QtGui import QDoubleValidator
from PyQt5.QtWidgets import QWidget, QFrame, QGridLayout, QVBoxLayout, \
                            QHBoxLayout, QGroupBox, QPushButton, QLineEdit, \
                            QDoubleSpinBox, QLabel, QScrollArea, QMessageBox, \
                            QCheckBox, QSplitter
# LiquidDiffract imports
from LiquidDiffract.gui import plot_widgets
from LiquidDiffract.gui import utility
from LiquidDiffract.core import data_utils
from LiquidDiffract.version import __appname__, __version__


class BkgUI(QWidget):

    # plots_changed is handled in main_widget.MainContainer and allows the
    # next tab to be updated
    plots_changed = pyqtSignal()
    # file_name_changed connects to main_widget.MainContainer.update_filename
    # this method is used to set the working directory based on the data_file
    # that is loaded
    file_name_changed = pyqtSignal()

    def __init__(self, parent):
        super(QWidget, self).__init__(parent)
        self.layout = QHBoxLayout(self)
        self.layout.setSpacing(0)

        # Make Config Widget
        self.bkg_config_widget = BkgConfigWidget()
        self.config_scroll_area = QScrollArea()
        self.config_scroll_area.setFrameShape(QFrame.NoFrame)
        self.config_scroll_area.setWidget(self.bkg_config_widget)
        self.config_scroll_area.setWidgetResizable(True)

        # Make vertical line separator
        self.vline = QFrame()
        self.vline.setFrameShape(QFrame.VLine)
        self.vline.setFrameShadow(QFrame.Sunken)
        self.vline.setObjectName("vline")

        # Make Plot Widget
        self.bkg_plot_widget = plot_widgets.BkgPlotWidget()
        self.plot_scroll_area = QScrollArea()
        self.plot_scroll_area.setWidget(self.bkg_plot_widget)
        self.plot_scroll_area.setWidgetResizable(True)
        self.plot_scroll_area.setFrameShape(QFrame.NoFrame)

        self.hsplitter = QSplitter(Qt.Horizontal)
        self.hsplitter.addWidget(self.config_scroll_area)
        self.hsplitter.addWidget(self.plot_scroll_area)
        self.hsplitter.setStretchFactor(0, 2)
        self.hsplitter.setStretchFactor(1, 5)
        self.layout.addWidget(self.hsplitter)

        self.setLayout(self.layout)

        self.data_file = os.path.abspath(os.getcwd())
        self.file_name_changed.emit()

        __a = np.asarray([])
        self.data = {'data_raw_x': __a, 'data_raw_y':   __a,
                     'data_x':     __a, 'data_y':       __a,
                     'bkg_raw_x':  __a, 'bkg_raw_y':    __a,
                     'bkg_x':      __a, 'bkg_y':        __a,
                     'bkg_y_sc':   __a, 'bkg_raw_y_sc': __a,
                     'cor_x':      __a, 'cor_y':        __a
                     }
        self.bkg_file = None
        self.data_file = None

        self.create_signals()

    def create_signals(self):
        self.bkg_config_widget.data_files_gb.load_data_btn.clicked.connect(self.load_data)
        self.bkg_config_widget.data_files_gb.load_bkg_btn.clicked.connect(self.load_bkg)
        self.bkg_config_widget.data_files_gb.plot_raw_check.stateChanged.connect(self.plot_data)
        self.bkg_config_widget.bkg_subtract_gb.bkg_sub_btn.clicked.connect(self.sub_bkg)
        self.bkg_config_widget.bkg_subtract_gb.scale_sb.valueChanged.connect(self.plot_data)
        self.bkg_config_widget.bkg_subtract_gb.toggled.connect(self.sub_bkg)
        self.bkg_config_widget.bkg_subtract_gb.auto_sc_btn.clicked.connect(self.auto_scale_bkg)
        self.bkg_config_widget.data_files_gb.dq_input.editingFinished.connect(self.dq_changed)

    def rebin_data(self, bkg='check'):
        '''
        Helper function to rebin background/data arrays when necessary.
        Because the S(Q) array is padded before fourier transform operations
        in core.core the size should be checked here. Checking takes place
        before rebinning as this could hang on very large files.
        '''
        try:
            _dx = np.float(self.bkg_config_widget.data_files_gb.dq_input.text())
        except ValueError:
            _dx = 0
        if bkg == 1:
            if self.data['bkg_raw_x'][-1] > ((2**self.fft_N / 2) * _dx):
                raise RuntimeWarning('Dataset size exceeds 2**N!')
            self.data['bkg_x'], self.data['bkg_y'] = data_utils.rebin_data(self.data['bkg_raw_x'], self.data['bkg_raw_y'], dx=_dx)
            self.plot_data()

        elif bkg == 0:
            if self.data['data_raw_x'][-1] > ((2**self.fft_N / 2) * _dx):
                 raise RuntimeWarning('Dataset size exceeds 2**N!')
            self.data['data_x'], self.data['data_y'] = data_utils.rebin_data(self.data['data_raw_x'], self.data['data_raw_y'], dx=_dx)
            # Delete any processed data when loading new file
            self.data['cor_x'] = np.asarray([])
            self.data['cor_y'] = np.asarray([])
            if not self.bkg_config_widget.bkg_subtract_gb.isChecked():
                self.sub_bkg()
            self.plots_changed.emit()
            self.plot_data()

        elif bkg == 'check':
            if self.data['bkg_raw_x'].size and self.bkg_file != None:
                 self.rebin_data(bkg=1)
            if self.data['data_raw_x'].size and self.data_file != None:
                 self.rebin_data(bkg=0)

    def load_data(self):
        __file_name = utility.get_filename(io='open')
        if __file_name:
            self.data_file = __file_name
        else:
            return
        try:
            self.data['data_raw_x'], self.data['data_raw_y'] = np.loadtxt(self.data_file, unpack=True)
            # Convert to angstroms if data_units == 1 (nm)
            if self.data_units:
                self.data['data_raw_x'] /= 10.0
        except ValueError:
            self.data_file = None
            self.load_file_error()
            return
        try:
            self.rebin_data(bkg=0)
        except RuntimeWarning:
             self.oversize_file_error()
             self.data_file = None
             return
        self.bkg_config_widget.data_files_gb.data_filename_lbl.setText(self.data_file.split('/')[-1])
        print(f'Data file: {self.data_file}')
        self.file_name_changed.emit()

    def load_bkg(self):
        __file_name = utility.get_filename(io='open', caption='Load Background File')
        if __file_name:
            self.bkg_file = __file_name
        else:
            return
        try:
            self.data['bkg_raw_x'], self.data['bkg_raw_y'] = np.loadtxt(self.bkg_file, unpack=True)
            if self.data_units:
                self.data['bkg_raw_x'] /= 10.0
        except ValueError:
            self.bkg_file = None
            self.load_file_error()
            return
        try:
            self.rebin_data(bkg=1)
        except RuntimeWarning:
            self.oversize_file_error()
            self.bkg_file = None
            return
        self.bkg_config_widget.data_files_gb.bkg_filename_lbl.setText(self.bkg_file.split('/')[-1])
        print(f'Background File: {self.bkg_file}')

    def plot_data(self):
        if self.data['bkg_y'].size:
            # First scale bkg data
            _bkg_scaling = self.bkg_config_widget.bkg_subtract_gb.scale_sb.value()
            self.data['bkg_raw_y_sc'] = self.data['bkg_raw_y'] * _bkg_scaling
            self.data['bkg_y_sc'] = self.data['bkg_y'] * _bkg_scaling
        if self.data['cor_x'].size:
            # Only re-subtract if already subtract button clicked
            if self.bkg_config_widget.bkg_subtract_gb.isChecked() and self.bkg_file is not None:
                # Subtract background from raw data if option chosen
                try:
                    self.data['cor_y'] = self.data['data_y'] - self.data['bkg_y_sc']
                except ValueError:
                    self.bkg_match_error()
                    return
            else:
                self.data['cor_y'] = self.data['data_y']
        if self.bkg_config_widget.data_files_gb.plot_raw_check.isChecked():
            _plot_raw = 1
        else:
            _plot_raw = 0
        self.bkg_plot_widget.update_plots(self.data, _plot_raw)
        # emit signal that data has changed to be picked up by second tab (optim_ui)
        self.plots_changed.emit()

    def sub_bkg(self):
        # Cor x = data x
        self.data['cor_x'] = self.data['data_x']
        self.plot_data()

    def auto_scale_bkg(self):
        if self.bkg_file is None:
            self.missing_bkg_file_error()
            return
        if len(self.data['data_y']) != len(self.data['bkg_y']):
            self.bkg_match_error()
            return
        bkg_scaling = minimize(data_utils.bkg_scaling_residual, 1,
                               args=(self.data['data_y'], self.data['bkg_y']),
                               method='nelder-mead',
                               options={'xtol': 1e-8, 'disp': False})
        bkg_scaling = bkg_scaling.x
        self.bkg_config_widget.bkg_subtract_gb.scale_sb.setValue(bkg_scaling)

    def dq_changed(self):
        try:
            self.rebin_data()
        except RuntimeWarning:
             self.dq_error()
             self.bkg_config_widget.data_files_gb.dq_input.setText('0.02')
             self.dq_changed()

    def load_file_error(self):
        message = ['Error loading file!', 'Unable to load file.\nPlease check filename,\nensure header lines are commented (#),\nand data is in Q-space']
        self.warning_message(message)

    def oversize_file_error(self):
        message = ['Error loading file!', 'Re-binned array size\nis too large! Check the data-units preferences setting, increase the Q-step or increase length of Fourier transform array (N)']
        self.warning_message(message)

    def missing_bkg_file_error(self):
        message = ['No background file!', 'Please load a background file']
        self.warning_message(message)
    
    def bkg_match_error(self):
        message = ['Error subtracting background!', 'Data and background\n do not match!']
        self.warning_message(message)

    def dq_error(self):
        message = ['Error setting Q-step!', 'Re-binned array size\nis too large! Increase the Q-step or length of Fourier transform array (N)']
        self.warning_message(message)

    def warning_message(self, message):
        self.error_msg = QMessageBox()
        self.error_msg.setIcon(QMessageBox.Warning)
        self.error_msg.setStandardButtons(QMessageBox.Ok)
        self.error_msg.setText(message[0])
        self.error_msg.setInformativeText((message[1]))
        self.error_msg.setWindowTitle(__appname__ + ' v' + __version__)
        self.error_msg.adjustSize()
        self.error_msg.show()


class BkgConfigWidget(QWidget):

    def __init__(self):
        super(BkgConfigWidget, self).__init__()

        self.vlayout = QVBoxLayout()
        self.vlayout.setContentsMargins(0, 0, 5, 0)
        self.vlayout.setSpacing(10)
        self.data_files_gb = DataFilesGroupBox()
        self.bkg_subtract_gb = BkgSubtractGroupBox()
        self.data_conv_gb = DataConvertGroupBox()
        self.vlayout.addWidget(self.data_files_gb, 1)
        self.vlayout.addWidget(self.bkg_subtract_gb, 1)
        self.vlayout.addWidget(self.data_conv_gb, 1)
        self.vlayout.addWidget(QWidget(), 3)
        self.setLayout(self.vlayout)


class DataFilesGroupBox(QGroupBox):

    def __init__(self, *args):
        super(DataFilesGroupBox, self).__init__(*args)
        self.setTitle('Data Files')
        self.setAlignment(Qt.AlignLeft)
        self.setStyleSheet('GroupBox::title{subcontrol-origin: margin; subcontrol-position: top left;}')
        
        self.create_widgets()
        self.style_widgets()
        self.create_layout()

    def create_widgets(self):
        self.load_data_btn = QPushButton("Load Data")
        self.data_filename_lbl = QLabel("None")

        self.load_bkg_btn = QPushButton("Load Background")
        self.bkg_filename_lbl = QLabel("None")

        self.data_lbl_frame = QScrollArea()
        self.bkg_lbl_frame = QScrollArea()

        self.plot_raw_check = QCheckBox()
        self.plot_raw_lbl = QLabel('Plot raw (unbinned) data?')
        self.plot_raw_check.setChecked(False)
        
        self.dq_label = QLabel('Re-binned resolution (Q-step):')
        self.dq_input = QLineEdit('0.02')

    def style_widgets(self):
        self.data_filename_lbl.setAlignment(Qt.AlignCenter)
        self.bkg_filename_lbl.setAlignment(Qt.AlignCenter)

        self.data_lbl_frame.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.data_lbl_frame.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.data_lbl_frame.setWidgetResizable(True)
        self.data_lbl_frame.setFrameShape(QFrame.NoFrame)

        self.bkg_lbl_frame.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.bkg_lbl_frame.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.bkg_lbl_frame.setWidgetResizable(True)
        self.bkg_lbl_frame.setFrameShape(QFrame.NoFrame)

        self.dq_label.setAlignment(Qt.AlignVCenter | Qt.AlignLeft)
        self.dq_input.setAlignment(Qt.AlignRight)
        self.dq_input.setMaximumWidth(60)
        self.dq_input.setValidator(QDoubleValidator())

    def create_layout(self):
        self.grid_layout = QGridLayout()
        self.grid_layout.setContentsMargins(25, 10, 25, 7)
        self.grid_layout.setSpacing(5)
        self.data_lbl_frame.setWidget(self.data_filename_lbl)
        self.bkg_lbl_frame.setWidget(self.bkg_filename_lbl)
        self.grid_layout.addWidget(self.load_data_btn, 0, 0)
        self.grid_layout.addWidget(self.data_lbl_frame, 0, 1)
        self.grid_layout.addWidget(self.load_bkg_btn, 1, 0)
        self.grid_layout.addWidget(self.bkg_lbl_frame, 1, 1)
        self.grid_layout.addWidget(self.plot_raw_lbl, 2, 0)
        self.grid_layout.addWidget(self.plot_raw_check, 2, 1)
        self.grid_layout.addWidget(self.dq_label, 3, 0)
        self.grid_layout.addWidget(self.dq_input)
        self.setLayout(self.grid_layout)


class BkgSubtractGroupBox(QGroupBox):

    def __init__(self, *args):
        super(BkgSubtractGroupBox, self).__init__(*args)
        self.setTitle('Subtract Bkg?')
        self.setAlignment(Qt.AlignLeft)
        self.setStyleSheet('GroupBox::title{subcontrol-origin: margin; subcontrol-position: top left;}')
        self.setCheckable(True)

        self.create_widgets()
        self.style_widgets()
        self.create_layout()
        self.create_signals()

    def create_widgets(self):

        self.scale_lbl = QLabel('Bkg Scaling: ')
        self.scale_sb = QDoubleSpinBox()
        self.scale_step = QLineEdit('0.01')
        self.auto_sc_btn = QPushButton('Auto-scale Bkg')
        self.bkg_sub_btn = QPushButton('Subtract Background')

    def style_widgets(self):

        self.scale_lbl.setAlignment(Qt.AlignVCenter | Qt.AlignRight)

        self.scale_sb.setValue(1.0)
        self.scale_sb.setSingleStep(0.01)
        self.scale_sb.setDecimals(3)
        self.scale_sb.setMinimumWidth(80)
        self.scale_sb.setAlignment(Qt.AlignRight)

        self.scale_step.setMaximumWidth(60)
        self.scale_step.setValidator(QDoubleValidator())
        self.scale_step.setAlignment(Qt.AlignRight)

    def create_layout(self):
        self.outer_layout = QVBoxLayout()
        self.outer_layout.setContentsMargins(25, 10, 25, 7)

        self.inner_layout = QHBoxLayout()
        self.inner_layout.setContentsMargins(0, 0, 0, 0)
        self.inner_layout.setSpacing(8)
        self.inner_layout.addWidget(self.scale_lbl, 0)
        self.inner_layout.addWidget(self.scale_sb, 1)
        self.inner_layout.addWidget(self.scale_step, 2)

        self.outer_layout.addLayout(self.inner_layout)
        self.outer_layout.addWidget(self.auto_sc_btn)
        self.outer_layout.addWidget(self.bkg_sub_btn)

        self.setLayout(self.outer_layout)

    def create_signals(self):
        self.scale_step.editingFinished.connect(self.scale_step_changed)

    def scale_step_changed(self):
        self.scale_sb.setSingleStep(float(str(self.scale_step.text())))


class DataConvertGroupBox(QGroupBox):

    def __init__(self, *args):
        super(DataConvertGroupBox, self).__init__(*args)
        self.setTitle('Convert 2-theta to Q-space')
        self.setAlignment(Qt.AlignLeft)
        self.setStyleSheet('GroupBox::title{subcontrol-origin: margin; subcontrol-position: top left;}')
        self.setCheckable(True)
        self.setChecked(False)

        self.data_file = None
        self.two_theta_data = None

        self.load_conv_data_btn = QPushButton('2 theta Data')
        self.data_filename_lbl = QLabel('None')

        self.lambda_lbl = QLabel('Lambda: ')
        self.lambda_input = QLineEdit('')
        self.lambda_input.setValidator(QDoubleValidator())
        self.lambda_input.setMaximumWidth(60)

        self.data_filename_lbl.setAlignment(Qt.AlignCenter)
        self.save_conv_data_btn = QPushButton('Convert')

        self.outer_layout = QVBoxLayout()
        self.outer_layout.setContentsMargins(25, 25, 25, 25)
        self.outer_layout.setSpacing(10)

        self.inner_layout = QHBoxLayout()
        self.inner_layout.setContentsMargins(0, 0, 0, 0)
        self.inner_layout.setSpacing(10)

        self.outer_layout.addWidget(self.load_conv_data_btn)
        self.outer_layout.addWidget(self.data_filename_lbl)

        self.inner_layout.addWidget(self.lambda_lbl)
        self.inner_layout.addWidget(self.lambda_input)

        self.outer_layout.addLayout(self.inner_layout)
        self.outer_layout.addWidget(self.save_conv_data_btn)

        self.setLayout(self.outer_layout)

        self.create_signals()

    def create_signals(self):
        self.load_conv_data_btn.clicked.connect(self.load_two_theta)
        self.save_conv_data_btn.clicked.connect(self.save_q_space)

    def load_two_theta(self):

        __file_name = utility.get_filename(io='open', caption='Load 2-theta data file')
        if not __file_name:
            return
        self.data_file = __file_name
        try:
            self.two_theta_data = np.loadtxt(self.data_file, unpack=False)
        except ValueError:
            _message = ['Error loading file!', 'Unable to load file.\nPlease check filename is correct and make sure header lines are commented (#)']
            self.error_msg = QMessageBox()
            self.error_msg.setIcon(QMessageBox.Warning)
            self.error_msg.setStandardButtons(QMessageBox.Ok)
            self.error_msg.setText(_message[0])
            self.error_msg.setInformativeText((_message[1]))
            self.error_msg.setWindowTitle(__appname__ + ' v' + __version__)
            self.error_msg.adjustSize()
            self.error_msg.show()
            self.data_file = None
            self.two_theta_data = None
            return
        self.data_filename_lbl.setText(self.data_file.split('/')[-1])
        __split_path = os.path.splitext(self.data_file)
        self.default_fname = __split_path[0] + '_qspace' + __split_path[1]

    def save_q_space(self):
        if self.data_file is None or self.two_theta_data is None:
            return
        try:
            __lambda = np.float(self.lambda_input.text())
        # Return if no wavelength set
        except ValueError:
            _message = ['No Wavelength Set!', 'Please set wavelength value']
            self.error_msg = QMessageBox()
            self.error_msg.setIcon(QMessageBox.Warning)
            self.error_msg.setStandardButtons(QMessageBox.Ok)
            self.error_msg.setText(_message[0])
            self.error_msg.setInformativeText((_message[1]))
            self.error_msg.setWindowTitle(__appname__ + ' v' + __version__)
            self.error_msg.adjustSize()
            self.error_msg.show()
            return
        try:
            # Get out filename
            self.convert_filename = utility.get_filename(io='save', caption='Save Q-space Data', directory=self.default_fname)
        except NameError:
            return
        if not self.convert_filename:
            return
        # Convert 2theta data
        __q_data = data_utils.convert_two_theta(self.two_theta_data[0], __lambda)
        __out_data = np.column_stack((__q_data, self.two_theta_data[1]))
        np.savetxt(self.convert_filename, __out_data)
        # Message user success
        self.success_msg = QMessageBox()
        self.success_msg.setIcon(QMessageBox.Information)
        self.success_msg.setStandardButtons(QMessageBox.Ok)
        self.success_msg.setText('Data converted to Q-Space!')
        self.success_msg.setInformativeText(('Lambda: ' + str(__lambda) + '\nConverted data: ' + self.convert_filename))
        self.success_msg.setWindowTitle(__appname__ + ' v' + __version__)
        self.success_msg.adjustSize()
        self.success_msg.show()
        # Clear variables
        del __lambda, __q_data, __out_data, self.convert_filename



# -*- coding: utf-8 -*-
__author__ = "Benedict J. Heinen"
__copyright__ = "Copyright 2018-2021, Benedict J. Heinen"
__email__ = "benedict.heinen@gmail.com"

from PyQt5.QtCore import Qt, pyqtSignal
from PyQt5.QtWidgets import QWidget, QVBoxLayout
import pyqtgraph as pg
import numpy as np
from LiquidDiffract.core import data_utils

pg_options = {'leftButtonPan': False, 'background': 0.9, 'foreground': 0.15,
              'antialias': True}
pg.setConfigOptions(**pg_options)


class BkgPlotWidget(QWidget):
    def __init__(self, *args, **kwargs):
        super(BkgPlotWidget, self).__init__(*args, **kwargs)

        self.layout = QVBoxLayout()
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setSpacing(8)
        self.create_plots()
        self.style_plots()

        self.create_signals()

        self.setLayout(self.layout)

    def create_plots(self):
        self.pg_layout_widget = pg.GraphicsLayoutWidget()
        self.pg_layout = pg.GraphicsLayout()
        self.pg_layout.setContentsMargins(0, 0, 0, 0)
        self.pg_layout_widget.setContentsMargins(0, 0, 0, 0)

        self.data_plot = CustomPlotItem()
        self.bkg_corrected_plot = CustomPlotItem()

        self.data_plot.plot(x=[], y=[], pen={'color': 0.1, 'width': 1.2})
        self.bkg_corrected_plot.plot(x=[], y=[], pen={'color': 0.1, 'width': 1.2})

        self.pg_layout.addItem(self.data_plot, row=1, col=0)
        self.pg_layout.addItem(self.bkg_corrected_plot, row=2, col=0)

        self.pg_layout_widget.addItem(self.pg_layout)

        self.layout.addWidget(self.pg_layout_widget)

    def style_plots(self):
        self.data_plot.setLabel('bottom', text='Q (1/A)')
        self.data_plot.setLabel('left', text='Intensity (a.u.)')

        self.bkg_corrected_plot.setLabel('bottom', text='Q (1/A)')
        self.bkg_corrected_plot.setLabel('left', text='Intensity (a.u.)')

        self.pos_label = pg.LabelItem(justify='right')
        self.pg_layout.addItem(self.pos_label, col=0, row=0)

    def update_plots(self, _data, _plot_raw):
        try:
            self.p1_a.clear()
            self.p2_a.clear()
            self.p3.clear()
        except AttributeError:
            pass
        try:
            self.p1_b.clear()
            self.p2_b.clear()
        except AttributeError:
            pass

        self.p1_a = self.data_plot.plot(x=_data['data_x'], y=_data['data_y'], pen={'color': 0.1, 'width': 1.2})
        self.p2_a = self.data_plot.plot(x=_data['bkg_x'], y=_data['bkg_y_sc'], pen={'color': '#342256', 'width': 1.2, 'style': Qt.DashLine})
        self.p3 = self.bkg_corrected_plot.plot(x=_data['cor_x'], y=_data['cor_y'], pen={'color': 0.1, 'width': 1.2})

        if _plot_raw:
            self.p1_b = self.data_plot.plot(x=_data['data_raw_x'], y=_data['data_raw_y'], pen=None, symbolPen={'color': 0.1}, symbolBrush=0.1, symbol='x', symbolSize=7)
            self.p2_b = self.data_plot.plot(x=_data['bkg_raw_x'], y=_data['bkg_raw_y_sc'], pen=None, symbolPen={'color': '#342256'}, symbolBrush='#342256', symbol='x', symbolSize=7)

        self.data_plot.vb.autoRange()
        self.bkg_corrected_plot.vb.autoRange()

    def create_signals(self):
        self.mouse_proxy = pg.SignalProxy(self.pg_layout.scene().sigMouseMoved, rateLimit=60, slot=self.mouse_moved)

    def mouse_moved(self, __evt):
        # using signal proxy turns original arguments into a tuple
        __pos = __evt[0]
        if self.data_plot.sceneBoundingRect().contains(__pos):
            __mousePoint = self.data_plot.vb.mapSceneToView(__pos)
            self.set_mouse_pos_label(__mousePoint)

            self.data_plot.vline.setPos(__mousePoint.x())
            self.data_plot.hline.setPos(__mousePoint.y())

            self.data_plot.vline.setPen((0, 135, 153), width=0.75)
            self.data_plot.hline.setPen((0, 135, 153), width=0.75)

            self.bkg_corrected_plot.vline.setPen(None)
            self.bkg_corrected_plot.hline.setPen(None)

        elif self.bkg_corrected_plot.sceneBoundingRect().contains(__pos):
            __mousePoint = self.bkg_corrected_plot.vb.mapSceneToView(__pos)
            self.set_mouse_pos_label(__mousePoint)

            self.bkg_corrected_plot.vline.setPos(__mousePoint.x())
            self.bkg_corrected_plot.hline.setPos(__mousePoint.y())

            self.bkg_corrected_plot.vline.setPen((0, 135, 153), width=0.75)
            self.bkg_corrected_plot.hline.setPen((0, 135, 153), width=0.75)

            self.data_plot.vline.setPen(None)
            self.data_plot.hline.setPen(None)

    def set_mouse_pos_label(self, pos):
        _pos_str = (f'<span style="font-size: 11pt; color:#008799">x='
                    f'{pos.x():.2f}, y={pos.y():.2f}</span'
                    )
        self.pos_label.setText(_pos_str)


class OptimPlotWidget(QWidget):

    def __init__(self, *args, **kwargs):
        super(OptimPlotWidget, self).__init__(*args, **kwargs)

        self.layout = QVBoxLayout()
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setSpacing(8)
        self.create_plots()
        self.style_plots()

        self.create_signals()

        self.setLayout(self.layout)

    def create_plots(self):
        self.pg_layout_widget = pg.GraphicsLayoutWidget()
        self.pg_layout = pg.GraphicsLayout()
        self.pg_layout.setContentsMargins(0, 0, 0, 0)
        self.pg_layout_widget.setContentsMargins(0, 0, 0, 0)

        self.data_plot = CustomPlotItem()
        self.iq_plot = CustomPlotItem()
        self.dr_plot = CustomPlotItem()

        self.data_plot.plot(x=[], y=[])
        self.iq_plot.plot(x=[], y=[])
        self.dr_plot.plot(x=[], y=[], pen={})

        self.pg_layout.addItem(self.data_plot, row=1, col=0)
        self.pg_layout.addItem(self.iq_plot, row=2, col=0)
        self.pg_layout.addItem(self.dr_plot, row=3, col=0)

        self.pg_layout_widget.addItem(self.pg_layout)

        self.layout.addWidget(self.pg_layout_widget)

    def style_plots(self):
        self.data_plot.setLabel('bottom', text='Q (1/A)')
        self.data_plot.setLabel('left', text='Intensity (a.u.)')

        self.iq_plot.setLabel('bottom', text='Q (1/A)')
        self.iq_plot.setLabel('left', text='i(Q)')

        self.dr_plot.setLabel('bottom', text='r (A)')
        self.dr_plot.setLabel('left', text='D(r)')

        self.pos_label = pg.LabelItem(justify='right')
        self.pg_layout.addItem(self.pos_label, col=0, row=0)

    def update_plots(self, _data):
        try:
            self.p1_a.clear()
            self.p2_a.clear()
            self.p3_a.clear()
        except AttributeError:
            pass
        try:
            self.p1_b.clear()
        except AttributeError:
            pass
        try:
            self.p2_b.clear()
            self.p3_b.clear()
        except AttributeError:
            pass
        try:
            self.p2_c.clear()
        except AttributeError:
            pass

        # Some versions of pyqtgraph cannot produce plot if nan values present
        # First value in some arrays is nan (e.g. int func)
        # For interference function this should be == 0 - S_inf
        # Fix nan values by interpolation
        if np.isnan(_data['int_func']).any():
             _data['int_func'] = data_utils.interp_nan(_data['int_func'])
        if np.isnan(_data['impr_int_func']).any():
             _data['impr_int_func'] = data_utils.interp_nan(_data['impr_int_func'])
        if np.isnan(_data['dr_x']).any():
             _data['dr_x'] = data_utils.interp_nan(_data['dr_x'])
        if np.isnan(_data['impr_dr_x']).any():
             _data['impr_dr_x'] = data_utils.interp_nan(_data['impr_dr_x'])

        # For the D(r) the r step = pi/q_max. Because the data is padded this
        # q_max is larger than the original q_max. i.e. q_max = dq * 2**N/2
        # Then: r_max = 2**N/2 * dr = pi/(dq * 2**N/2) * 2**N/2 = pi/dq
        # Displaying the full length data is not necessary as most of the 
        # useful information is contained at fairly low r, and the length of r
        # is controlled only by sampling frequency in Q-space. At high r-values
        # the D(r) is dominated by ripples from the truncated integral (0-qmax)
        # The max value of r with 'real' q resolution is 1/dq.
        _window = 0
        try:
            _dq = _data['iq_x'][1] - _data['iq_x'][0]
            try:
                _window = np.argmax(_data['dr_x'] >= (1/_dq))
            except ValueError:
                 pass
        except IndexError:
            pass

        self.p1_a = self.data_plot.plot(x=_data['cor_x_cut'], y=_data['cor_y_cut'], pen={'color': 0.1, 'width': 1.2})
        self.p1_b = self.data_plot.plot(x=_data['cor_x_cut'], y=_data['rescaled_cor_y_cut'], pen={'color': '#342256', 'width': 1.2, 'style': Qt.DashLine})
        self.p2_a = self.iq_plot.plot(x=_data['iq_x'], y=_data['int_func'], pen={'color': 0.1, 'width': 1.2})
        self.p2_b = self.iq_plot.plot(x=_data['impr_iq_x'], y=_data['impr_int_func'], pen={'color': '#342256', 'width': 1.2, 'style': Qt.DashLine})
        self.p3_a = self.dr_plot.plot(x=_data['dr_x'][:_window], y=_data['dr_y'][:_window], pen={'color': 0.1, 'width': 1.2})
        self.p3_b = self.dr_plot.plot(x=_data['impr_dr_x'][:_window], y=_data['impr_dr_y'][:_window], pen={'color': '#342256', 'width': 1.2, 'style': Qt.DashLine})

        if _data['mod_func'] != 'None':
            # Apply modification function and handle any nans that arise 
            _modified_int_func = _data['modification'] * _data['int_func']
            if np.isnan(_modified_int_func).any():
                 _modified_int_func = data_utils.interp_nan(_modified_int_func)
            self.p2_c = self.iq_plot.plot(x=_data['iq_x'], y=_modified_int_func, pen={'color': '#342256', 'width': 0.8, 'style': Qt.DashLine})

        self.data_plot.vb.autoRange()
        self.iq_plot.vb.autoRange()
        self.dr_plot.vb.autoRange()

    def create_signals(self):
        self.mouse_proxy = pg.SignalProxy(self.pg_layout.scene().sigMouseMoved, rateLimit=60, slot=self.mouse_moved)

    def mouse_moved(self, __evt):
        # using signal proxy turns original arguments into a tuple
        __pos = __evt[0]
        if self.data_plot.sceneBoundingRect().contains(__pos):
            __mousePoint = self.data_plot.vb.mapSceneToView(__pos)
            self.set_mouse_pos_label(__mousePoint)

            self.data_plot.vline.setPos(__mousePoint.x())
            self.data_plot.hline.setPos(__mousePoint.y())

            self.data_plot.vline.setPen((0, 135, 153), width=0.75)
            self.data_plot.hline.setPen((0, 135, 153), width=0.75)

            self.iq_plot.vline.setPen(None)
            self.iq_plot.hline.setPen(None)
            self.dr_plot.vline.setPen(None)
            self.dr_plot.hline.setPen(None)

        elif self.iq_plot.sceneBoundingRect().contains(__pos):
            __mousePoint = self.iq_plot.vb.mapSceneToView(__pos)
            self.set_mouse_pos_label(__mousePoint)

            self.iq_plot.vline.setPos(__mousePoint.x())
            self.iq_plot.hline.setPos(__mousePoint.y())

            self.iq_plot.vline.setPen((0, 135, 153), width=0.75)
            self.iq_plot.hline.setPen((0, 135, 153), width=0.75)

            self.data_plot.vline.setPen(None)
            self.data_plot.hline.setPen(None)
            self.dr_plot.vline.setPen(None)
            self.dr_plot.hline.setPen(None)

        elif self.dr_plot.sceneBoundingRect().contains(__pos):
            __mousePoint = self.dr_plot.vb.mapSceneToView(__pos)
            self.set_mouse_pos_label(__mousePoint)

            self.dr_plot.vline.setPos(__mousePoint.x())
            self.dr_plot.hline.setPos(__mousePoint.y())

            self.dr_plot.vline.setPen((0, 135, 153), width=0.75)
            self.dr_plot.hline.setPen((0, 135, 153), width=0.75)

            self.data_plot.vline.setPen(None)
            self.data_plot.hline.setPen(None)
            self.iq_plot.hline.setPen(None)
            self.iq_plot.vline.setPen(None)

    def set_mouse_pos_label(self, pos):
        _pos_str = (f'<span style="font-size: 11pt; color:#008799">x='
                    f'{pos.x():.2f}, y={pos.y():.2f}</span'
                    )
        self.pos_label.setText(_pos_str)


class ResultsPlotWidget(QWidget):

    def __init__(self, *args, **kwargs):
        super(ResultsPlotWidget, self).__init__(*args, **kwargs)

        self.layout = QVBoxLayout()
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setSpacing(8)
        self.create_plots()
        self.style_plots()

        self.create_signals()

        self.setLayout(self.layout)

    def create_plots(self):
        self.pg_layout_widget = pg.GraphicsLayoutWidget()
        self.pg_layout = pg.GraphicsLayout()
        self.pg_layout.setContentsMargins(0, 0, 0, 0)
        self.pg_layout_widget.setContentsMargins(0, 0, 0, 0)

        self.sq_plot = CustomPlotItem()
        self.gr_plot = WindowedPlotItem()
        self.rdf_plot = WindowedPlotItem()

        self.sq_plot.plot(x=[], y=[])
        self.gr_plot.plot(x=[], y=[])
        self.rdf_plot.plot(x=[], y=[])
        self.rdf_plot.setXLink(self.gr_plot)

        self.pg_layout.addItem(self.sq_plot, row=1, col=0)
        self.pg_layout.addItem(self.gr_plot, row=2, col=0)
        self.pg_layout.addItem(self.rdf_plot, row=3, col=0)

        self.pg_layout_widget.addItem(self.pg_layout)

        self.layout.addWidget(self.pg_layout_widget)

    def style_plots(self):
        self.sq_plot.setLabel('bottom', text='Q (1/A)')
        self.sq_plot.setLabel('left', text='S(Q)')

        self.gr_plot.setLabel('bottom', text='r (A)')
        self.gr_plot.setLabel('left', text='g(r)')

        self.rdf_plot.setLabel('bottom', text='r (A)')
        self.rdf_plot.setLabel('left', text='RDF(r)')

        self.pos_label = pg.LabelItem(justify='right')
        self.pg_layout.addItem(self.pos_label, col=0, row=0)

    def update_plots(self, _data):
        try:
            self.p1.clear()
            self.p2.clear()
            self.p3.clear()
        except AttributeError:
            pass
        try:
            self.p1_mod.clear()
        except AttributeError:
            pass

        # Some versions of pyqtgraph cannot produce plot if nan values present
        # Fix nan values by interpolation
        if np.isnan(_data['sq_y']).any():
             _data['sq_y'] = data_utils.interp_nan(_data['sq_y'])
        if np.isnan(_data['gr_y']).any():
             _data['gr_y'] = data_utils.interp_nan(_data['gr_y'])
        if np.isnan(_data['rdf_y']).any():
             _data['rdf_y'] = data_utils.interp_nan(_data['rdf_y'])
        if np.isnan(_data['sq_y_mod']).any():
            _data['sq_y_mod'] = data_utils.interp_nan(_data['sq_y_mod'])
        _window = 0
        # Determine data window for Q-space resolution, dq
        try:
            _dq = _data['sq_x'][1] - _data['sq_x'][0]
            try:
                _window = np.argmax(_data['gr_x'] >= (1/_dq))
            except ValueError:
                 pass
        except IndexError:
            pass

        self.p1 = self.sq_plot.plot(x=_data['sq_x'], y=_data['sq_y'], pen={'color': 0.1, 'width': 1.2})
        if _data['mod_func'] != None:
            self.p1_mod = self.sq_plot.plot(x=_data['sq_x'], y=_data['sq_y_mod'], pen={'color': '#342256', 'width': 0.8, 'style': Qt.DashLine})
        self.p2 = self.gr_plot.plot(x=_data['gr_x'][:_window], y=_data['gr_y'][:_window], pen={'color': 0.1, 'width': 1.2})
        self.p3 = self.rdf_plot.plot(x=_data['rdf_x'][:_window], y=_data['rdf_y'][:_window], pen={'color': 0.1, 'width': 1.2})

        # Limit the inital view to important information
        try:
            self.x_max = _data['sq_x'][-1]
        except IndexError:
            return

        _gr_cut = np.nan_to_num(_data['gr_y'][np.where(_data['gr_x'] < self.x_max)])
        self.y_min_gr = np.min(_gr_cut)
        self.y_max_gr = np.max(_gr_cut)

        _rdf_cut = np.nan_to_num(_data['rdf_y'][np.where(_data['rdf_x'] < self.x_max)])
        self.y_min_rdf = np.min(_rdf_cut)
        self.y_max_rdf = np.max(_rdf_cut)

        self.set_gr_window()
        self.set_rdf_window()

        self.sq_plot.vb.autoRange()

    def set_gr_window(self):
        try:
            self.gr_plot.vb.setRange(xRange=(0, self.x_max),
                                     yRange=(self.y_min_gr, self.y_max_gr))
        except:
            return

    def set_rdf_window(self):
        try:
            self.rdf_plot.vb.setRange(xRange=(0, self.x_max),
                                      yRange=(self.y_min_rdf, self.y_max_rdf))
        except:
            return

    def create_signals(self):
        self.mouse_proxy = pg.SignalProxy(self.pg_layout.scene().sigMouseMoved, rateLimit=60, slot=self.mouse_moved)
        self.gr_plot.reset_window.connect(self.set_gr_window)
        self.rdf_plot.reset_window.connect(self.set_rdf_window)

    def mouse_moved(self, __evt):
        # Using signal proxy turns original args into tuple
        __pos = __evt[0]
        if self.sq_plot.sceneBoundingRect().contains(__pos):
            __mousePoint = self.sq_plot.vb.mapSceneToView(__pos)
            self.set_mouse_pos_label(__mousePoint)

            self.sq_plot.vline.setPos(__mousePoint.x())
            self.sq_plot.hline.setPos(__mousePoint.y())

            self.sq_plot.vline.setPen((0, 135, 153), width=0.75)
            self.sq_plot.hline.setPen((0, 135, 153), width=0.75)

            self.gr_plot.vline.setPen(None)
            self.gr_plot.hline.setPen(None)
            self.rdf_plot.vline.setPen(None)
            self.rdf_plot.hline.setPen(None)

        elif self.gr_plot.sceneBoundingRect().contains(__pos):
            __mousePoint = self.gr_plot.vb.mapSceneToView(__pos)
            self.set_mouse_pos_label(__mousePoint)

            self.gr_plot.vline.setPos(__mousePoint.x())
            self.gr_plot.hline.setPos(__mousePoint.y())

            self.gr_plot.vline.setPen((0, 135, 153), width=0.75)
            self.gr_plot.hline.setPen((0, 135, 153), width=0.75)

            self.sq_plot.vline.setPen(None)
            self.sq_plot.hline.setPen(None)
            self.rdf_plot.vline.setPen(None)
            self.rdf_plot.hline.setPen(None)

        elif self.rdf_plot.sceneBoundingRect().contains(__pos):
            __mousePoint = self.rdf_plot.vb.mapSceneToView(__pos)
            self.set_mouse_pos_label(__mousePoint)

            self.rdf_plot.vline.setPos(__mousePoint.x())
            self.rdf_plot.hline.setPos(__mousePoint.y())

            self.rdf_plot.vline.setPen((0, 135, 153), width=0.75)
            self.rdf_plot.hline.setPen((0, 135, 153), width=0.75)

            self.sq_plot.vline.setPen(None)
            self.sq_plot.hline.setPen(None)
            self.gr_plot.hline.setPen(None)
            self.gr_plot.vline.setPen(None)

    def set_mouse_pos_label(self, pos):
        _pos_str = (f'<span style="font-size: 11pt; color:#008799">x='
                    f'{pos.x():.2f}, y={pos.y():.2f}</span'
                    )
        self.pos_label.setText(_pos_str)


class StructurePlotWidget(QWidget):

    def __init__(self, *args, **kwargs):
        super(StructurePlotWidget, self).__init__(*args, **kwargs)

        self.layout = QVBoxLayout()
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setSpacing(8)
        self.create_plots()
        self.style_plots()

        self.create_signals()

        self.setLayout(self.layout)

    def create_plots(self):

        # Seperate widgets for integrating RDF, int Tr, and curve fitting
        # Widget for integrating across RDF
        self.rdf_int_layout_widget = pg.GraphicsLayoutWidget()
        self.rdf_int_layout_widget.setContentsMargins(0, 0, 0, 0)
        self.rdf_int_layout = pg.GraphicsLayout()
        self.rdf_int_layout.setContentsMargins(0, 0, 0, 0)
        # Widget for integrating T(r)
        # Separate widget used as easier than handling all the children
        self.tr_int_layout_widget = pg.GraphicsLayoutWidget()
        self.tr_int_layout_widget.setContentsMargins(0, 0, 0, 0)
        self.tr_int_layout = pg.GraphicsLayout()
        self.tr_int_layout.setContentsMargins(0, 0, 0, 0)

        # Curve fitting widget
        # One widget for RDF and Tr and only data and label change - all the
        # user added curves stay the same on switch
        self.fit_layout_widget = pg.GraphicsLayoutWidget()
        self.fit_layout_widget.setContentsMargins(0, 0, 0, 0)
        self.fit_layout = pg.GraphicsLayout()
        self.fit_layout.setContentsMargins(0, 0, 0, 0)
        self.fit_layout.setSpacing(0)

        # RDF(r) plot
        self.rdf_plot = WindowedPlotItem()
        self.rdf_plot.plot(x=[], y=[])
        self.rdf_int_layout.addItem(self.rdf_plot, row=1, col=0)
        self.rdf_int_layout_widget.addItem(self.rdf_int_layout)

        # T(r) plot
        self.tr_plot = WindowedPlotItem()
        self.tr_plot.plot(x=[], y=[])
        self.tr_int_layout.addItem(self.tr_plot, row=1, col=0)
        self.tr_int_layout_widget.addItem(self.tr_int_layout)

        # Curve (Gaussian) fitting plot
        self.fit_plot = WindowedPlotItem()
        self.fit_plot.plot(x=[], y=[])

        # Create fit range selection tool
        self.fit_limits = pg.LinearRegionItem(values=(0, 10), orientation='vertical', span=(0, 1), movable=True, swapMode='block', brush=pg.mkBrush(None), hoverBrush=pg.mkBrush(None))
        self.deselect_lower = pg.LinearRegionItem(values=(0, 0), orientation='vertical', span=(0, 1), movable=False, swapMode='block', pen=pg.mkPen(None), brush=pg.mkBrush((229,229,229,205)), hoverPen=pg.mkPen(None), hoverBrush=None)
        self.deselect_upper = pg.LinearRegionItem(values=(10, 10), orientation='vertical', span=(0, 1), movable=False, swapMode='block', pen=pg.mkPen(None), brush=pg.mkBrush((229,229,229,205)), hoverPen=pg.mkPen(None), hoverBrush=None)
        self.fit_plot.addItem(self.fit_limits)
        self.fit_plot.addItem(self.deselect_lower)
        self.fit_plot.addItem(self.deselect_upper)

        # Fitting residuals plot
        self.res_plot = WindowedPlotItem()
        self.res_plot.plot(x=[], y=[])
        self.res_plot.setXLink(self.fit_plot)
        self.fit_layout.addItem(self.fit_plot, row=1, col=0)
        self.fit_layout.addItem(self.res_plot, row=2, col=0)
        self.fit_layout.layout.setRowStretchFactor(1, 3)
        self.fit_layout.layout.setRowStretchFactor(2, 2)
        self.fit_layout_widget.addItem(self.fit_layout)

        # Add widgets to the layout
        self.layout.addWidget(self.rdf_int_layout_widget)
        self.layout.addWidget(self.tr_int_layout_widget)
        self.layout.addWidget(self.fit_layout_widget)

        # Set RDF(r) plot (integration) as default view
        self.rdf_int_layout_widget.setVisible(True)
        self.tr_int_layout_widget.setVisible(False)
        self.fit_layout_widget.setVisible(False)

    def style_plots(self):

        self.rdf_plot.setLabel('bottom', text='r (A)')
        self.rdf_plot.setLabel('left', text='RDF(r)')

        self.tr_plot.setLabel('bottom', text='r (A)')
        self.tr_plot.setLabel('left', text='T(r)')

        # No label on fit_plot - just on residuals
        self.fit_plot.setLabel('left', text='RDF(r)')
        self.res_plot.setLabel('bottom', text='r (A)')
        self.res_plot.setLabel('left', text='residuals')
        # Fix axis width so plots are aligned
        self.fit_plot.getAxis('bottom').setStyle(showValues=False)
        self.fit_plot.getAxis('left').setWidth(55)
        self.res_plot.getAxis('left').setWidth(55)

        self.rdf_xaxis = pg.InfiniteLine(pos=0, angle=0, movable=False, pen={'color': 'k', 'width': 0.75})
        self.rdf_plot.addItem(self.rdf_xaxis)
        self.tr_xaxis = pg.InfiniteLine(pos=0, angle=0, movable=False, pen={'color': 'k', 'width': 0.75})
        self.tr_plot.addItem(self.tr_xaxis)
        self.fit_xaxis = pg.InfiniteLine(pos=0, angle=0, movable=False, pen={'color': 'k', 'width': 0.75})
        self.fit_plot.addItem(self.fit_xaxis)
        self.res_xaxis = pg.InfiniteLine(pos=0, angle=0, movable=False, pen={'color': 'k', 'width': 0.75})
        self.res_plot.addItem(self.res_xaxis)
        self.fit_xaxis.setZValue(3)

        # Make fit_limits region invisible to start
        self.toggle_fit_limits(False)

        # Add position label to all plot views
        self.pos_label_rdf = pg.LabelItem(justify='right')
        self.rdf_int_layout.addItem(self.pos_label_rdf, col=0, row=0)

        self.pos_label_tr = pg.LabelItem(justify='right')
        self.tr_int_layout.addItem(self.pos_label_tr, col=0, row=0)

        self.pos_label_fit = pg.LabelItem(justify='right')
        self.fit_layout.addItem(self.pos_label_fit, col=0, row=0)

        self.rdf_plot.vline.setPen((0, 135, 153), width=0.75)
        self.rdf_plot.hline.setPen((0, 135, 153), width=0.75)

        self.tr_plot.vline.setPen((0, 135, 153), width=0.75)
        self.tr_plot.hline.setPen((0, 135, 153), width=0.75)

        self.fit_plot.vline.setPen((0, 135, 153), width=0.75)
        self.fit_plot.hline.setPen((0, 135, 153), width=0.75)
        self.res_plot.vline.setPen((0, 135, 153), width=0.75)
        self.res_plot.hline.setPen((0, 135, 153), width=0.75)

        # Set Z-value of selection region to plot above curves
        self.deselect_lower.setZValue(1)
        self.deselect_upper.setZValue(1)
        self.fit_limits.setZValue(2)
        self.fit_plot.vline.setZValue(6)
        self.fit_plot.hline.setZValue(6)
        self.fit_plot.getAxis('bottom').setZValue(7)

        # Colors for gaussian curves
        self.gauss_colors = [(26,121,199), (22,171,69), (246,174,45),
                             (242,100,25), (99,18,153), (5,117,59),
                             (196,35,72), (90,73,57)]

    def toggle_fit_limits(self, toggle_bool, obj_x=None, sq_x=None):
        self.fit_limits.setVisible(toggle_bool)
        self.deselect_lower.setVisible(toggle_bool)
        self.deselect_upper.setVisible(toggle_bool)
        if toggle_bool:
            # Set fit_limits default region and bounds to new data
            self.deselect_lower.setRegion([0,0])
            self.deselect_upper.setRegion([sq_x[-1],obj_x[-1]])
            self.fit_limits.setRegion([obj_x[0],sq_x[-1]])
            self.fit_limits.setBounds([0, obj_x[-1]])

    def clear_gauss_curves(self):
        try:
            self.p_gauss_model.clear()
            self.p_res.clear()
            self.p_res_full.clear()
            # Clear peak curves
            for _p_peak in self.p_peaks:
                _p_peak.clear()
                del _p_peak
            self.p_peaks = []
            # Delete peak labels
            for _peak_label in self.p_peaks_labels:
                _peak_label.deleteLater()
                del _peak_label
            self.p_peaks_labels = []
            # Delete arrows (pg.ArrowItem has no deleteLater method)
            for _peak_arrow in self.p_peaks_arrows:
                self.fit_plot.removeItem(_peak_arrow)
                del _peak_arrow
            self.p_peaks_arrows = []

        except AttributeError:
            pass

    def clear_plots(self, _clear_all=False):
        try:
            self.p_rdf.clear()
            self.p_tr.clear()
            self.p_fit.clear()
        except AttributeError:
            pass

        self.clear_gauss_curves()

        try:
            self.p_Na.clear()
        except AttributeError:
            pass

        try:
            self.p_Nb.clear()
        except AttributeError:
            pass

        try:
            self.p_Nc.clear()
        except AttributeError:
            pass

        if _clear_all:
            self.toggle_fit_limits(False)
            try:
                self.rdf_plot.removeItem(self.r0_line_rdf)
                self.tr_plot.removeItem(self.r0_line_tr)
                self.rdf_plot.removeItem(self.rpmax_line_rdf)
                self.tr_plot.removeItem(self.rpmax_line_tr)
                self.rdf_plot.removeItem(self.rmax_line_rdf)
                self.tr_plot.removeItem(self.rmax_line_tr)
                self.rdf_plot.removeItem(self.rmin_line_rdf)
                self.tr_plot.removeItem(self.rmin_line_rdf)

                self.r0_line_rdf.deleteLater()
                self.r0_line_tr.deleteLater()
                self.rpmax_line_rdf.deleteLater()
                self.rpmax_line_tr.deleteLater()
                self.rmax_line_rdf.deleteLater()
                self.rmax_line_tr.deleteLater()
                self.rmin_line_rdf.deleteLater()
                self.rmin_line_rdf.deleteLater()

                del self.r0_line_rdf
                del self.r0_line_tr
                del self.rpmax_line_rdf
                del self.rpmax_line_tr
                del self.rmax_line_rdf
                del self.rmax_line_tr
                del self.rmin_line_rdf
                del self.rmin_line_rdf
            except AttributeError:
                pass


    def update_plots(self, _data):

        # First clear the plots
        self.clear_plots()

        # Some versions of pyqtgraph cannot produce plot if nan values present
        # Fix nan values by interpolation
        if np.isnan(_data['rdf_y']).any():
             _data['rdf_y'] = data_utils.interp_nan(_data['rdf_y'])
        if np.isnan(_data['tr_y']).any():
             _data['tr_y'] = data_utils.interp_nan(_data['tr_y'])

        # Interpolate data for smoother plots
        # Ignoring for now
        #_data['rdf_x'], _data['rdf_y'] = data_utils.rebin_data(_data['rdf_x'], _data['rdf_y'], dx=0.01)
        #_data['tr_x'], _data['tr_y'] = data_utils.rebin_data(_data['tr_x'], _data['tr_y'], dx=0.01)

        # Create areas of N integrals
        _Na_area_idx = np.where((_data['tr_x'] > _data['r0']) & (_data['tr_x'] < _data['rpmax']))        
        _Nb_area_idx = np.where((_data['rdf_x'] > _data['r0']) & (_data['rdf_x'] < _data['rmax']))
        _Nc_area_idx = np.where((_data['rdf_x'] > _data['r0']) & (_data['rdf_x'] < _data['rmin']))

        _r0_tr_pt = data_utils.interp_data(_data['tr_x'], _data['tr_y'], _data['r0'])
        _r0_rdf_pt = data_utils.interp_data(_data['rdf_x'], _data['rdf_y'], _data['r0'])
        _rpmax_tr_pt = data_utils.interp_data(_data['tr_x'], _data['tr_y'], _data['rpmax'])        
        _rmax_rdf_pt = data_utils.interp_data(_data['rdf_x'], _data['rdf_y'], _data['rmax'])
        _rmin_rdf_pt = data_utils.interp_data(_data['rdf_x'], _data['rdf_y'], _data['rmin'])

        _Na_area_x = np.concatenate(([_data['r0']], _data['tr_x'][_Na_area_idx], [_data['rpmax']]))
        _Na_area_y = np.concatenate((_r0_tr_pt, _data['tr_y'][_Na_area_idx], _rpmax_tr_pt))
        _Nb_area_x = np.concatenate(([_data['r0']], _data['rdf_x'][_Nb_area_idx], [_data['rmax']]))
        _Nb_area_y = np.concatenate((_r0_rdf_pt, _data['rdf_y'][_Nb_area_idx], _rmax_rdf_pt))
        _Nc_area_x = np.concatenate(([_data['r0']], _data['rdf_x'][_Nc_area_idx], [_data['rmin']]))
        _Nc_area_y = np.concatenate((_r0_rdf_pt, _data['rdf_y'][_Nc_area_idx], _rmin_rdf_pt))

        # Plot fill areas for integrals behind function/data curve
        # Nc is plotted first/behind as it extends beyond Nb
        # Only plot if r0 < r
        if _data['r0'] < _data['rmin']:
            self.p_Nc = self.rdf_plot.plot(x=_Nc_area_x, y=_Nc_area_y, brush=(26,121,199, 245), pen=None, fillLevel=0)
        if _data['r0'] < _data['rpmax']:
            self.p_Na = self.tr_plot.plot(x=_Na_area_x, y=_Na_area_y, brush=(199,26,74, 245), pen=None, fillLevel=0)
        if _data['r0'] < _data['rmax']:
            self.p_Nb = self.rdf_plot.plot(x=_Nb_area_x, y=_Nb_area_y, brush=(199,26,74, 245), pen=None, fillLevel=0)

        # Plot data/functions RDF(r), T(r)
        self.p_rdf = self.rdf_plot.plot(x=_data['rdf_x'], y=_data['rdf_y'], pen={'color': 0.1, 'width': 1.2})
        self.p_tr = self.tr_plot.plot(x=_data['tr_x'], y=_data['tr_y'], pen={'color': 0.1, 'width': 1.2})
        self.p_fit = self.fit_plot.plot(x=_data['rdf_x'], y=_data['obj_fun'], pen={'color': 0.1, 'width': 1.2})

        # Limit the inital view to important information
        try:
            self.x_max = _data['sq_x'][-1]
        except IndexError:
            return

        # Plot integral limits as movable InfLines 
        try:
            # Set line positions
            self.r0_line_rdf.setPos(_data['r0'])
            self.r0_line_tr.setPos(_data['r0'])

            self.rpmax_line_rdf.setPos(_data['rpmax'])
            self.rpmax_line_tr.setPos(_data['rpmax'])

            self.rmax_line_rdf.setPos(_data['rmax'])
            self.rmax_line_tr.setPos(_data['rmax'])

            self.rmin_line_rdf.setPos(_data['rmin'])
            self.rmin_line_tr.setPos(_data['rmin'])

        except AttributeError:
            # Initial creation of lines
            # Create draggable lines for r0, rmax, rpmax and rmin
            self.r0_line_rdf = pg.InfiniteLine(pos=_data['r0'], movable=True, label='r0', bounds=[0, self.x_max], 
                                               labelOpts={'position': 0.75, 'movable': True, 'color': (0,10,40)}, name='r0')
            self.r0_line_tr = pg.InfiniteLine(pos=_data['r0'], movable=True, label='r0', bounds=[0, self.x_max], 
                                              labelOpts={'position': 0.75, 'movable': True, 'color': (0,10,40)}, name='r0')

            self.rpmax_line_rdf = pg.InfiniteLine(pos=_data['rpmax'], movable=True, label='r\'max', bounds=[0, self.x_max], 
                                                  labelOpts={'position': 0.8, 'movable': True, 'color': (0,10,40)}, name='rpmax')
            self.rpmax_line_tr = pg.InfiniteLine(pos=_data['rpmax'], movable=True, label='r\'max', bounds=[0, self.x_max], 
                                                 labelOpts={'position': 0.8, 'movable': True, 'color': (0,10,40)}, name='rpmax')

            self.rmax_line_rdf = pg.InfiniteLine(pos=_data['rmax'], movable=True, label='rmax', bounds=[0, self.x_max], 
                                                 labelOpts={'position': 0.85, 'movable': True, 'color': (0,10,40)}, name='rmax')
            self.rmax_line_tr = pg.InfiniteLine(pos=_data['rmax'], movable=True, label='rmax', bounds=[0, self.x_max], 
                                                labelOpts={'position': 0.85, 'movable': True, 'color': (0,10,40)}, name='rmax')

            self.rmin_line_rdf = pg.InfiniteLine(pos=_data['rmin'], movable=True, label='rmin', bounds=[0, self.x_max], 
                                                 labelOpts={'position': 0.75, 'movable': True, 'color': (0,10,40)}, name='rmin')
            self.rmin_line_tr = pg.InfiniteLine(pos=_data['rmin'], movable=True, label='rmin', bounds=[0, self.x_max], 
                                                labelOpts={'position': 0.75, 'movable': True, 'color': (0,10,40)}, name='rmin')

            self.rdf_plot.addItem(self.r0_line_rdf)
            self.tr_plot.addItem(self.r0_line_tr)

            self.rdf_plot.addItem(self.rpmax_line_rdf)
            self.tr_plot.addItem(self.rpmax_line_tr)

            self.rdf_plot.addItem(self.rmax_line_rdf)
            self.tr_plot.addItem(self.rmax_line_tr)

            self.rdf_plot.addItem(self.rmin_line_rdf)
            self.tr_plot.addItem(self.rmin_line_tr)        

        if _data['gauss_model'].size:
            # Plot individual gaussian peaks
            self.p_peaks = []
            self.p_peaks_labels = []
            self.p_peaks_arrows = []
            for i, _peak_model in enumerate(_data['gauss_peaks']):
                # Get color - use i%8 to cycle list of colors
                _color = self.gauss_colors[i%len(self.gauss_colors)]
                _p_peak = self.fit_plot.plot(x=_data['rdf_x'], y=_peak_model, pen={'color': _color, 'width': 1.2})
                _p_peak.setZValue(4)
                self.p_peaks.append(_p_peak)
                # Label peak at location of curve maximum
                _max_idx = np.argmax(_peak_model)
                _peak_loc = (_data['rdf_x'][_max_idx], _peak_model[_max_idx])
                _peak_arrow = pg.ArrowItem(pos=_peak_loc, angle=260, pen=_color, brush=_color, tailLen=30, tailWidth=2, headLen=10, headWidth=4, pxMode=True)
                _peak_label = pg.TextItem(text=_data['gauss_peaks_names'][i], color=_color, anchor=(1.0,2.47))
                _peak_label.setPos(*_peak_loc)
                self.fit_plot.addItem(_peak_label)
                self.fit_plot.addItem(_peak_arrow)
                self.p_peaks_labels.append(_peak_label)
                self.p_peaks_arrows.append(_peak_arrow)

            # Plot gaussian model (sum of individual peaks)
            # change color of this
            self.p_gauss_model = self.fit_plot.plot(x=_data['rdf_x'], y=_data['gauss_model'], pen={'color': (199,26,74), 'width': 1.8, 'style':Qt.DashLine})
            self.p_gauss_model.setZValue(5)
            # Plot residuals over whole r range in grey
            self.p_res_full = self.res_plot.plot(x=_data['rdf_x'], y=_data['gauss_residuals_full'], pen={'color': 0.75, 'width': 1.2})
            # Plot residuals over fit range in colour
            self.p_res = self.res_plot.plot(x=_data['fit_r'], y=_data['gauss_residuals'], pen={'color': (26,121,199), 'width': 1.5})

        # Force repaint
        self.rdf_int_layout.update()
        self.tr_int_layout.update()
        self.fit_layout.update()

    def update_plot_windows(self, _data, _reset_view):
        _rdf_cut = np.nan_to_num(_data['rdf_y'][np.where(_data['rdf_x'] < self.x_max)])
        self.y_min_rdf = np.min(_rdf_cut)
        self.y_max_rdf = np.max(_rdf_cut)

        _tr_cut = np.nan_to_num(_data['tr_y'][np.where(_data['tr_x'] < self.x_max)])
        self.y_min_tr = np.min(_tr_cut)
        self.y_max_tr = np.max(_tr_cut)

        _obj_fun_cut = np.nan_to_num(_data['obj_fun'][np.where(_data['rdf_x'] < self.x_max)])
        self.y_min_fit = np.min(_obj_fun_cut)
        self.y_max_fit = np.max(_obj_fun_cut)

        if _reset_view:
            self.set_rdf_window()
            self.set_tr_window()
            self.set_fit_window()
            self.set_res_window(_data)

    def set_rdf_window(self):
        try:
            self.rdf_plot.vb.setRange(xRange=(0, self.x_max),
                                      yRange=(self.y_min_rdf, self.y_max_rdf))
        except:
            return

    def set_tr_window(self):
        try:
            self.tr_plot.vb.setRange(xRange=(0, self.x_max),
                                     yRange=(self.y_min_tr, self.y_max_tr))
        except:
            return

    def set_fit_window(self):
        try:
            self.fit_plot.vb.setRange(xRange=(0, self.x_max),
                                     yRange=(self.y_min_fit, self.y_max_fit))
        except:
            return

    def set_res_window(self, _data):
        try:
            self.y_min_res = np.min(_data['gauss_residuals'])
            self.y_max_res = np.max(_data['gauss_residuals'])
            self.res_plot.vb.setRange(yRange=(self.y_min_res, self.y_max_res))
        except:
            return

    def create_signals(self):
        self.rdf_mouse_proxy = pg.SignalProxy(self.rdf_int_layout.scene().sigMouseMoved, rateLimit=60, slot=self.mouse_moved_rdf)
        self.tr_mouse_proxy = pg.SignalProxy(self.tr_int_layout.scene().sigMouseMoved, rateLimit=60, slot=self.mouse_moved_tr)
        self.gauss_mouse_proxy = pg.SignalProxy(self.fit_layout.scene().sigMouseMoved, rateLimit=60, slot=self.mouse_moved_fit)

        self.rdf_plot.reset_window.connect(self.set_rdf_window)
        self.tr_plot.reset_window.connect(self.set_tr_window)
        self.fit_plot.reset_window.connect(self.set_fit_window)
        self.res_plot.reset_window.connect(self.set_fit_window)

        self.fit_limits.sigRegionChanged.connect(self.update_linear_region)

    def update_linear_region(self):
        _lb, _ub = self.fit_limits.getRegion()
        _min, _ = self.deselect_lower.getRegion()
        _, _max = self.deselect_upper.getRegion()
        self.deselect_lower.setRegion((_min, _lb))
        self.deselect_upper.setRegion((_ub, _max))

    def mouse_moved_rdf(self, __evt):
        # Use different slots as mouse always in scene bounding rectangle
        # Using signal proxy turns original args into tuple
        __pos = __evt[0]
        __mousePoint = self.rdf_plot.vb.mapSceneToView(__pos)
        self.set_mouse_pos_label(__mousePoint)
        self.rdf_plot.vline.setPos(__mousePoint.x())
        self.rdf_plot.hline.setPos(__mousePoint.y())

    def mouse_moved_tr(self, __evt):
        __pos = __evt[0]
        __mousePoint = self.tr_plot.vb.mapSceneToView(__pos)
        self.set_mouse_pos_label(__mousePoint)
        self.tr_plot.vline.setPos(__mousePoint.x())
        self.tr_plot.hline.setPos(__mousePoint.y())

    def mouse_moved_fit(self, __evt):
        __pos = __evt[0]
        if self.fit_plot.sceneBoundingRect().contains(__pos):
            __mousePoint = self.fit_plot.vb.mapSceneToView(__pos)
            self.set_mouse_pos_label(__mousePoint)
            self.fit_plot.vline.setPos(__mousePoint.x())
            self.fit_plot.hline.setPos(__mousePoint.y())
            self.fit_plot.vline.setPen((0, 135, 153), width=0.75)
            self.fit_plot.hline.setPen((0, 135, 153), width=0.75)
            self.res_plot.vline.setPen(None)
            self.res_plot.hline.setPen(None)

        elif self.res_plot.sceneBoundingRect().contains(__pos):
            __mousePoint = self.res_plot.vb.mapSceneToView(__pos)
            self.set_mouse_pos_label(__mousePoint)
            self.res_plot.vline.setPos(__mousePoint.x())
            self.res_plot.hline.setPos(__mousePoint.y())
            self.res_plot.vline.setPen((0, 135, 153), width=0.75)
            self.res_plot.hline.setPen((0, 135, 153), width=0.75)
            self.fit_plot.vline.setPen(None)
            self.fit_plot.hline.setPen(None)

    def set_mouse_pos_label(self, pos):
        _pos_str = (f'<span style="font-size: 11pt; color:#008799">x='
                    f'{pos.x():.2f}, y={pos.y():.2f}</span'
                    )
        self.pos_label_rdf.setText(_pos_str)
        self.pos_label_tr.setText(_pos_str)
        self.pos_label_fit.setText(_pos_str)


class CustomPlotItem(pg.PlotItem):

    def __init__(self, *args, **kwargs):
        super(CustomPlotItem, self).__init__(*args, **kwargs)

        self.enableAutoRange()
        self.setMenuEnabled(False)
        self.vline = pg.InfiniteLine(angle=90, movable=False)
        self.hline = pg.InfiniteLine(angle=0, movable=False)

        self.vline.setPen(width=0)
        self.hline.setPen(width=0)

        self.addItem(self.vline, ignoreBounds=True)
        self.addItem(self.hline, ignoreBounds=True)

        self.vb.mouseDoubleClickEvent = self.mouse_double_click_event

    def mouse_double_click_event(self, __evt):
        if __evt.button() == Qt.RightButton:
            self.vb.autoRange()


class WindowedPlotItem(CustomPlotItem):

    reset_window = pyqtSignal()

    def autoBtnClicked(self):
        self.reset_window.emit()



# -*- coding: utf-8 -*-
"""
Objective and utility functions for peak fitting in LiquidDiffract
"""
__author__ = "Benedict J Heinen"
__copyright__ = "Copyright 2021, Benedict J. Heinen"
__email__ = "benedict.heinen@gmail.com"

from math import sqrt, pi
from functools import lru_cache
import numpy as np
from scipy.special import erf

@lru_cache(maxsize=2)
def cached_sqrt(root):
    '''Cache sqrt(2pi) and sqrt(2). Very minor speed bost'''
    if root == '2':
        return sqrt(2)
    else:
        return sqrt(2*pi)


def skew_gauss(r, N_ab, r_ab, sigma_ab, xi_ab, W_ab, c_b):
    '''
    Gaussian type function from Cormier, 2019 (pg. 1053) 
    
    Args:
        r - r-values to evaluate the function over (np array)
        N_ab* - Avg. coordination number for pair alpha-beta
        r_ab* - Avg. bond length for pair alpha-beta
        sigma_ab* - Bond length distribution
        xi_ab* - Skewness parameter (set to zero for pure gaussian)
        W_ab - X-ray pair correlation weighting for pair alpha-beta
        c_b - Atomic fraction of beta in composition
        
        *fitted parameter
    
    Returns:
        skew_gauss_r - skew_gauss(r)
    '''
    r_diff = r - r_ab
    scaling_factor = (W_ab * N_ab) / (c_b * sigma_ab * cached_sqrt('2pi'))
    exp_term = np.exp(-r_diff**2/(2 * sigma_ab**2))
    skew = 1 + erf(xi_ab * r_diff / (sigma_ab * cached_sqrt('2')))
    skew_gauss_r = scaling_factor * exp_term * skew
    return skew_gauss_r


def gauss_obj_func(params, r, data=None, **kwargs):
    '''
    lmfit objective function to evaluate an arbitrary number of gaussian-type
    peaks (gauss_func(r)) against data
    
    Args:
        
        params - lmfit.Parameters() object containing values, bounds, and
                 refinement options for parameters [N, r, s] for each peak
        
        r - r values to evaluate over
        
        data - f(r) values to fit model to (optional) 
               if None: returns model, else: returns model - data
        
        kwargs - dict containing W_ab, and c_b for each peak, along with
                 peak_list
                 
        kwargs['peak_list'] must be a list of unique peak ids. These are used
        to determine corresponding params/kwargs for each peak. The naming 
        convention for these is 'key'+str(peak_id). i.e. for peak id 0, params
        are N0, r0, s0, xi0, etc. kwargs are W0, c_b0, etc.
    
    Returns:
        
        model (if data == None) - sum of gaussian peaks evaluated over r
        
        residual (if data != None) - difference between model and data

    '''
    # Unpack lmfit parameters into dictionary
    param_values = params.valuesdict()
    # Get list of uniqe peak indices
    peak_idx_list = kwargs['peak_idx_list']
    # Initiate list to store each individual peak evaluated over r
    model = []
    # Loop over peak_idx_list and evaluate gauss_func for each peak
    for peak_idx in peak_idx_list:
        # Get id string
        peak_idx_str = str(peak_idx)
        # Unpack args for clarity
        # (r,) N_ab, r_ab, sigma_ab, W_ab, c_b (, sqrt_2pi)
        N_ab = param_values['N'+peak_idx_str]
        r_ab = param_values['r'+peak_idx_str]
        s_ab = param_values['s'+peak_idx_str]
        xi_ab = param_values['xi'+peak_idx_str]
        W_ab = kwargs['W'+peak_idx_str]
        c_b = kwargs['c_b'+peak_idx_str]
        # r_ab and s_ab may be zero if lower-bounds are 0
        # Catch these cases here as otherwise resulting NaNs cause 
        # lmfit to fail
        if r_ab == 0:
            r_ab += 1.e-6
        if s_ab == 0:
            s_ab += 1.e-6
        peak_model = skew_gauss(r, N_ab, r_ab, s_ab, xi_ab, W_ab, c_b)
        # Debugging
        if np.isnan(peak_model).any():
            print('nans here (internal) in one of the gauss peaks!!')
        model.append(peak_model)
    # Convet to np.array
    model = np.asarray(model)
    # SUM gaussians = RDF(r) - divide gaussian by r to fit to T(r)
    if kwargs['active_func'] == 'tr':
        # Only divide where r!=0
        model = np.divide(model, r, out=np.zeros_like(model, dtype=np.float64), where=r!=0)
    # TODO - Convolve gaussian peak models with step function
    # Sum peaks
    model = np.sum(model, axis=0)
    
    if data is None:
        return model
    else:
        return model - data



'''
LiquidDiffract
==============

LiquidDiffract is a graphical data analysis application for processing
experimental total X-ray diffraction data of liquids and disordered solids.

LiquidDiffract can perform background subtraction,fourier transformation,
numerical optimisation/normalisation of the quantitative structure factor,
and estimation of sample density.

The sub-package LiquidDiffract.core provides useful functions for common
numerical operations on total diffraction data.

Core functions available from LiquidDiffract.core.core

    >>> import LiquidDiffract.core.core
    >>> help(LiquidDiffract.core.core)

Additional utilities from LiquidDiffract.core.data_utils

    >>> import LiquidDiffract.core.data_utils
    >>> help(LiquidDiffract.core.data_utils)


Please see the documentation on the LiquidDiffract homepage for more details.
<https://www.github.com/bjheinen/LiquidDiffract>

Copyright (c) 2019 Benedict J. Heinen

Licensed under the GNU General Public License (GPL), version 3 or later.

'''
from LiquidDiffract.version import __version__



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
'''
This script computes the function X^2(n, rho).
It does a simple brute force calculation of X^2 across the range of (n, rho),
where n is the number of iterations in the Eggert procedure and rho is the
density.

'''
__author__ = 'Benedict J. Heinen'
__copyright__ = 'Copyright 2019, Benedict J. Heinen'
__license__ = 'GNU GPL v3'
__email__ = 'benedict.heinen@gmail.com'


import numpy as np
import warnings
warnings.filterwarnings('ignore')
import matplotlib.pyplot as plt
from scipy.interpolate import UnivariateSpline
import LiquidDiffract.core.core as liquid
from LiquidDiffract.core.data_utils import rebin_data
import time

def update_progress_bar(iteration_number, total_iterations, elapsed_time, estimated_total_time):
    '''
    Prints a progress bar, with % completed and estimated time remaining, to the terminal.
    '''
    percentage_done = 100 * (iteration_number / total_iterations)
    if iteration_number == total_iterations: 
        percentage_done = 100
    percent_str = ('{0:.1f}').format(percentage_done)
    # Char length of progress bar = 50
    bar_length = 50
    fill_char = '█'
    fill_length = bar_length * iteration_number // total_iterations
    progress_bar_str = fill_char * fill_length + '-' * (bar_length - fill_length)
    eta = estimated_total_time - elapsed_time
    if eta < 60:
        eta_str = str(int(eta)) + 's' + ' '*12
    else:
        eta_str = time.strftime('%H:%M:%S', time.gmtime(eta))
    if iteration_number == total_iterations:
        eta_str = '0s'
    if total_iterations == 1:
        return
    print('\rProgress: |%s| %s%% Complete  |  ETA: %s' % (progress_bar_str, percent_str, eta_str), end='\r')


def brute_compute(brute_array, func_args, 
                  brute_iterations, start_time, estimated_total_time,
                  slice_number = 0):
    '''
    Brute force optimisation, computing CHI^2 at each value in 'brute_array'.
    Warning!: Uses recursion!
        This allows the function to be used for brute force calculations
        that include a 3rd variable - e.g. background scaling, or r_min.
        But watch out for python's recursion limit!
    '''
    if np.shape(brute_array)[0] == 2:
        outer = brute_array[0]
        inner = brute_array[1]
        res = []
        for slice_number, n_iter in enumerate(outer):
            func_args['n_iter'] = n_iter
            brute_slice = brute_compute(inner, func_args, brute_iterations,
                                        start_time, estimated_total_time,
                                        slice_number=slice_number)
            res.append(brute_slice)
        return np.asarray(res)
    else:
        pass
    func_args = tuple(func_args.values())
    res = []
    for iteration, rho in enumerate(brute_array):
        chi_sq = liquid.refinement_objfun(rho, *func_args)
        elapsed_time = time.time() - start_time
        iteration_num = (iteration + 1) + (len(brute_array)*slice_number)
        update_progress_bar(iteration_num, brute_iterations, elapsed_time, estimated_total_time)
        res.append(chi_sq)
    res = np.asarray(res)
    return res

# Data Options
data_file = 'example_data.dat'
q_max = 11.8
comp = {'Ga':(31,0,1)}
dq = 0.02
sq_form = 'faber-ziman'
mod_func = None
window_start = None
r_min = 2.3
fft_N = 12

# Set the array of values to compute X^2
brute_rho_array = np.arange(0.01, 0.08, 0.001)
brute_eggert_iterations_array = np.arange(1, 50, 1)

# Load data
x, y = np.loadtxt(data_file, unpack=True)
# rebin with set dq (steps of q values)
x, y = rebin_data(x, y, dx=dq)
# Cut data at Q_max
x, y = x[np.where(x < q_max)], y[np.where(x < q_max)]

func_args = {'Q_data': x, 'I_data': y, 'comp': comp,
             'r_min': r_min, 'n_iter': int(np.mean(brute_eggert_iterations_array)),
             'method': sq_form, 'mod_func': mod_func,
             'window_start': window_start, 'fft_N': fft_N,
             'opt_rho': 1, 'opt_bkg': 0}

brute_array = (brute_eggert_iterations_array, brute_rho_array)
brute_iterations = len(brute_rho_array) * len(brute_eggert_iterations_array)
# Here we calculate the time to make 1 iteration
t_start = time.time()
brute_compute([1], func_args, 1, 1, 1, 1)
t_end = time.time()
iteration_time = t_end - t_start
# Estimate total time to complete calculations in seconds
estimated_total_time = (iteration_time * brute_iterations)
start_time = time.time()
# Calculate Chi^2 at all values of rho
brute_res = brute_compute(brute_array, func_args,
                          brute_iterations, start_time, estimated_total_time)

brute_res = np.nan_to_num(brute_res)
np.save('brute_res.npy', brute_res, allow_pickle=True)
np.save('brute_rho.npy', brute_rho_array, allow_pickle=True)
np.save('brute_niter.npy', brute_eggert_iterations_array, allow_pickle=True)

fig1 = plt.figure()
for i in range(len(brute_res)):
    plt.plot(brute_rho_array, brute_res[i], label=('n_iter = ' + str(brute_eggert_iterations_array[i])))
plt.legend()
plt.xlabel(r'$\rho$')
plt.ylabel(r'$\chi^2$')
plt.show()
#plt.savefig('fitting region spread.svg')

map_x, map_y = np.meshgrid(brute_rho_array, brute_eggert_iterations_array)
brute_res_normalised = []
for niter_slice in brute_res:
     brute_res_normalised.append(niter_slice/np.min(niter_slice))
brute_res_normalised = np.asarray(brute_res_normalised)

log_map = np.log(brute_res_normalised)
mask = np.where(log_map == 0)
dbl_log_map = log_map
dbl_log_map[mask] = 1e-12
dbl_log_map = np.log(dbl_log_map)

fig2 = plt.figure()
opt_density = []
for index, niter in enumerate(brute_eggert_iterations_array):
    opt_density.append(brute_rho_array[np.where(dbl_log_map[index] == np.min(dbl_log_map[index]))][0])
opt_density = np.asarray(opt_density)
plt.plot(brute_eggert_iterations_array, 1/opt_density)
plt.xlabel('Number of iterations in refinement procedure')
plt.ylabel(r'$\rho^{-1}$')
plt.show()
#plt.savefig('refined density vs n iterations.png', dpi=1200)


fig3 = plt.figure()
# Plot the minimum points
cp = plt.contourf(map_x, map_y, dbl_log_map, 50)
plt.title(r'Minimum $\chi^2$ values')
plt.ylabel('No. iterations in refinement procedure')
plt.xlabel(r'Refined $\rho_0$ $(atoms/\AA^3)$')
plt.show()
#plt.savefig('rho_vs_iterations', dpi=2400)

# Calculate curvature at each Chi^2 minimum
second_derivatives = []
for res in brute_res:
    rho_min = brute_rho_array[np.where(res == np.min(res))]
    spline = UnivariateSpline(brute_rho_array, res/1e6)
    second_derivatives.append(spline(rho_min))
second_derivatives = np.asarray(second_derivatives)

# Plot d^2chi^2 / drho^2 vs n_iter
fig4, ax1 = plt.subplots()
ax1.plot(brute_eggert_iterations_array, second_derivatives)
plt.title('Second derivative at $\chi^2$ minima vs No. of iterations')
plt.xlabel('Number of iterations in refinement procedure')
plt.ylabel(r'$\frac{d^2\chi^2}{d\rho^2}$ $at$ $\rho{}={}\rho_{min}$')
ax2 = fig4.add_axes([0.35, 0.45, 0.3, 0.2])
ax2.plot(brute_eggert_iterations_array[3:17], second_derivatives[3:17])
plt.xticks(brute_eggert_iterations_array[3:17][::2])
plt.show()
#plt.savefig('second_derivatives', dpi=2400)


# -*- coding: utf-8 -*-
"""
Utility functions for common data operations in LiquidDiffract
"""
__author__ = "Benedict J Heinen"
__copyright__ = "Copyright 2018-2021, Benedict J. Heinen"
__email__ = "benedict.heinen@gmail.com"

from functools import lru_cache, wraps
import numpy as np
import scipy.interpolate
import scipy.optimize
import scipy.signal

def rebin_data(x, y, dx=0.02, x_lim=None):
    '''
    Re-bins x,y data (using interpolation) to achieve a set step size of x (dx)
    Extrapolates x < x_min to zero
    '''
    if x_lim != None:
        x_rebin = np.arange(x_lim[0], x_lim[1], dx)
    else:
        x_rebin = np.arange(0, x[-1], dx)
    f_interp = scipy.interpolate.interp1d(x, y,
                                          kind='cubic',
                                          fill_value='extrapolate')
    y_rebin = f_interp(x_rebin)
    return x_rebin, y_rebin

def interp_data(x, y, x2):
    '''
    Helper function to return interpolated y-value at given x2 for f(x)=y
    '''
    f_interp = scipy.interpolate.interp1d(x, y, kind='cubic',
                                          fill_value='extrapolate')
    return f_interp(x2).flatten()


def convert_two_theta(two_theta, wavelength):
    '''Convert 2theta values to Q-space values at a known wavelength'''
    q_data = (4 * np.pi / wavelength) * np.sin(np.radians(two_theta)/2)
    return q_data


def zero_norm(y, shift=None):
    '''Shift data on y axis. If no 'shift' given, first value is set to zero'''
    if shift == None:
         shift = y[0]
    return y - shift


def interp_nan(y):
     '''
     Uses interpolation to fix nan values in an array
     This is useful for nan values at start of interference function array
     which cannot be set to zero with np.nan_to_num
     '''
     # Array of True/False values for nan location
     nans = np.isnan(y)
     # Return nonzero (non-nan) values
     f = lambda z: z.nonzero()[0]
     y[nans] = np.interp(f(nans), f(~nans), y[~nans])
     return y


def bkg_scaling_residual(bkg_scaling, *args):
    '''
    Returns the mean deviation between data and a scaled background

    Uses *args syntax so the function can be passed to a solver

    It is expected that *args == (data, bkg)

    Assumes data/bkg x-axis values are the same

    Arguments:

    bkg_scaling -- Background scaling factor

    data -- y values only

    bkg -- unscaled background data (y values only)

    '''
    (data, bkg) = args
    return np.mean(np.abs(data - (bkg*bkg_scaling)))


def smooth_data(data, method='savitzky-golay', window_length=31, poly_order=3):
    '''
    Applies a savitzky-golay filter to smooth noise from x, y data

    Arguments:

    data -- Must be a 1d array of y data only.

    window_length -- The length of the filter window (number of coefficients)
                     Must be a positive odd integer

    poly_order -- The order of the polynomial used to fit the samples
                  This must be an integer less than window_length.
    '''
    if method == 'savitzky-golay':
        return scipy.signal.savgol_filter(data, window_length, poly_order)


def find_integration_limits(r, rdf, rho=None, peak_search_limit=10.0, search_method='first'):
    '''
    Computes the integration limits used in the calculation of the 
    first coordination number for a monatomic system.

    The function returns several values for use in different methods of
    calculating the coordination number. 
    See 'core.integrate_coordination_sphere' for more information on
    the methods implemented in LiquidDiffract.

    The location of the peak corresponding to the first coordination sphere 
    (the 1st peak in RDF(r)) is first approximated from the discrete data 
    using a peak finding algorithm. A continuous function is then generated
    using cubic interpolation and the position of r_0 refined using a root
    finding algorithm and the positions of r_max, rp_max, and r_min are 
    refined through calls to scipy.minimize

    Args:
        r   - r values (numpy array)
        rdf - Corresponding values of the radial distribution function, RDF(r) (numpy array)

        rho - optional kwarg to provide number density (rho). This is used to
              calculate g(r) if the peak search in rdf(r) fails

        peak_search_limit - optional kwarg to set limit on intial peak search
                            this is useful as the most prominent (topographic)
                            peak is taken and the function can be confused by
                            choosing a peak at high r. If the main peak is the
                            nth peak, there must be at least n+1 peaks found

    Returns:
        r_0 - Leading edge of the 1st peak in RDF(r), taken as the nearest root
        rp_max - Peak centre in T(r) (equivalent to centre in r*g(r))
        r_max - Peak centre in RDF(r) (equivalent to centre in r^2*g(r))
        r_min - Position of 1st minimum after the 1st peak in RDF(r)
    '''
    # Calculate T(r)
    with np.errstate(divide='ignore', invalid='ignore'):
        Tr = np.nan_to_num(rdf/r)

    # Create continuous functions of RDF(r) and T(r) via cubic interpolation
    rdf_interp = scipy.interpolate.interp1d(r, rdf, kind='cubic', fill_value='extrapolate')
    Tr_interp = scipy.interpolate.interp1d(r, Tr, kind='cubic', fill_value='extrapolate')

    # Take the last root in RDF(r) as the leading edge of the 1st peak, r_0
    # Uses Brent method to find a zero of RDF(r) on the sign changing interval [a , b].
    # First find last sign-changing interval in discrete data
    brent_a = np.argwhere(np.sign(rdf) == -1)[-1]
    # Find root using scipy.optimize.brentq
    r_0 = scipy.optimize.brentq(rdf_interp, r[brent_a], r[brent_a+1])

    # Generate list of indices of approximate peak positions in the discrete 
    # data. Here using scipy.signal but using np.diff(np.sign(np.diff(rdf)) to
    # avoid the extra import may be less expensive
    peak_list, _ = scipy.signal.find_peaks(rdf[r<peak_search_limit])

    # Select 1st peak after r0 as 1st coordination sphere
    peak_idx_first = np.argmax(peak_list>brent_a)

    # Select most prominent peak as 1st coordination sphere and get positions
    # at the base either side
    peak_prom, left_bases, right_bases = scipy.signal.peak_prominences(rdf, peak_list)
    peak_idx_prom = np.argmax(peak_prom)
    
    if search_method == 'first':
        peak_idx = peak_idx_first
    elif search_method == 'prominent':
        peak_idx = peak_idx_prom
    else:
        raise AttributeError('\'method\' must be \'first\' or \'prominent\'')
    
    peak_bases = (r[left_bases[peak_idx]], r[right_bases[peak_idx]])

    # Get approximate position of next peak to provide upper bound on r_min
    try:
        next_peak = r[peak_list[peak_idx+1]]
    # Handle IndexError if chosen peak is last in peak_list
    except IndexError:
        # Re-do peak search in g(r) if density (rho) available
        if rho != None:
            with np.errstate(divide='ignore', invalid='ignore'):
                gr = rdf/(4*np.pi*rho*r**2)
            gr_peak_list, _ = scipy.signal.find_peak(gr[r<peak_search_limit])
            # Take the first peak after r0
            next_peak = r[gr_peak_list[np.argmax(gr_peak_list>brent_a)+1]]
        # If rho not available use
        else:
            next_peak = peak_search_limit

    # Refine rp_max (peak centre in r g(r))
    rp_max = scipy.optimize.minimize_scalar(lambda x: -Tr_interp(x), method='bounded', bounds=peak_bases).x
    # Refine r_max (peak centre in r^2 g(r))
    r_max = scipy.optimize.minimize_scalar(lambda x: -rdf_interp(x), method='bounded', bounds=peak_bases).x

    # Refine r_min (position of 1st minimum after 1st peak)
    # Note: r_min should be global minimum but optimisation may return local min
    r_min = scipy.optimize.minimize_scalar(rdf_interp, method='bounded', bounds=(r_max, next_peak)).x

    return r_0, rp_max, r_max, r_min


def data_cache(*args, **kwargs):
    '''
    Provides a decorator to cache the result of functions of the form
    f(hashable_type, 1d_Array). The numpy array is converted to a tuple so
    that functools.lru_cache can be used.

    Here it is used to cache the computation of element/Q specific form-
    factors and compton scattering.
    '''
    def decorator(function):
        @wraps(function)
        def wrapper(hashable_arg, np_array):
            hashable_array = tuple(np_array)
            return cached_wrapper(hashable_arg, hashable_array)

        @lru_cache(*args, **kwargs)
        def cached_wrapper(hashable_arg, hashable_array):
            array = np.array(hashable_array)
            return function(hashable_arg, array)

        wrapper.cache_info = cached_wrapper.cache_info
        wrapper.cache_clear = cached_wrapper.cache_clear

        return wrapper
    return decorator


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This script shows an example usage of the LiquidDiffract library. Whilst LiquidDiffract
was primarily designed as a graphical application, the functions in its core module can
be easily called from a python script. This is useful for batch processing files (where
an ongoing graphical view of the steps is not neccessary), or computing probability 
density functions from pre-existing structure factors (fourier transformation). It can 
also be used to call functions that are intermediary calculation steps in the graphical
application - e.g. retreiving atomic form factors or calculating compton scattering 
intensities for an element or multi-component composition.

The two primary functions in normal usage are:

    LiquidDiffract.core.core.calc_structure_factor
    LiquidDiffract.core.core.calc_impr_interference_func

calc_structure_factor computes the structure factor, S(Q), for given sample
composition, density, and experimentally measured sample scattering data.

calc_impr_interference_func optimises the interference function, S(Q) - S_inf,
based on the iterative method developed by Eggert et al., 2002, Phy Rev B, 65(17).

See <https://github.com/bjheinen/LiquidDiffract> for more details.
"""
__author__ = 'Benedict J. Heinen'
__copyright__ = 'Copyright 2018-2022, Benedict J. Heinen'
__email__ = 'benedict.heinen@gmail.com'
__license__='GPLv3'

# Import required python modules
import numpy as np
# Import optional python modules used in this example
import matplotlib.pyplot as plt

from scipy.optimize import minimize
# import the LiquidDiffract core module
import LiquidDiffract.core.core as liquid
import LiquidDiffract.core.data_utils as data_utils

# First set composition
# The composition should be a dictionary dictionary with short element names as keys,
# and values as tuples of the form (Z, charge, number_of_atoms)
composition = {'Ga': (31,0,1)}
# Set initial density - in atoms per cubic Angstrom
rho = 0.055

# Load data
# LiquidDiffract expects x data to be Q values in Angstroms (not nm!)
# The file 'example_data.dat' has already been background corrected
q_raw, I_raw = np.loadtxt('example_data.dat', unpack=True, skiprows=0)

# Next rebin data and trim if necessary
# Re-bin via interpolation so dq (steps of q) is consistent to allow FT
# This can be done using the rebin_data function from core.data_utils
dq=0.02
q_data, I_data = data_utils.rebin_data(q_raw, I_raw, dx=dq, x_lim=[q_raw[0], 11.8])

# Apply any other data treatment
# e.g. a savitsky-golay filter to smooth the data
I_data = data_utils.smooth_data(I_data, method='savitzky-golay', 
                                window_length=31, poly_order=3)

# First calculate the interference function i(Q)
# i(Q) = S(Q) - S_inf (S_inf = 1 for monatomic samples, or in the Faber-Ziman formalism)
structure_factor = liquid.calc_structure_factor(q_data, I_data, composition, rho, method='faber-ziman')
interference_func = structure_factor - liquid.calc_S_inf(composition, q_data, method='faber-ziman')

# store original interference function
interference_func_0 = interference_func

# A low r region cutoff must be chosen for the refinement method of Eggert et al. (2002)
# A density value in atoms(or molecules) / A^3 must also be set. This will be
# used as the starting estimate if the density is refined.
r_min = 2.3
rho_0 = rho

# The function 'calc_impr_interference_func' calculates an improved estimate of
# the interference function via the iterative procedure described in Eggert et al., 2002.
#
# calc_impr_interference_func requires the following arguments:
# q_data - q values
# I_data / i(Q) - the treated intensity data or interference function (depends on opt_flag used)
# composition - composition dictionary
# r_min - low r cut-off
# iter_limit - iteration limit for Eggert procedure
# method - method of calculating S(Q) ('ashcroft-langreth' and 'faber-ziman' are currently supported)
# mod_func - modification function to use ('None', 'Cosine-window' or 'Lorch')
# window_start - window_start of cosine function (if in use)
# fft_N sets the size of the array during the fourier transform to 2**N
# opt_flag - set to 0 if not running from solver
#
# The last positional argument (opt_flag) signals that you only want the chi_squared value to be returned.
# This is useful if using a solver to refine the density by minimising the chi_squared value

# e.g.
# Return the improved i(Q) & chi^2 at density = rho_0
iter_limit = 25
method = 'faber-ziman'
mod_func = 'Cosine-window'
window_start = 7
fft_N = 12
args = (q_data, interference_func_0, composition, rho_0,
        r_min, iter_limit, method, mod_func, window_start, fft_N)
# Store the refined interference function at rho_0
interference_func_1, chi_sq_1 = liquid.calc_impr_interference_func(*args)

# Next we pass the function calc_impr_interference_func to
# a solver to estimate the density. We use a separate objective function
# core.redinement_objfun to make this simpler.
# The argument opt_rho is set to 1 to indicate we are refining the density
opt_rho = 1
# The argument opt_bkg is set to 0 to indicate we are not refining
# the background sclaing factor.
opt_bkg = 0
# When refinind the density an iter_limit <= 10 is recommended
iter_limit_refine = 7
args = (q_data, I_data, composition, r_min,
        iter_limit_refine, method,
        mod_func, window_start, fft_N,
        opt_rho, opt_bkg)
# Set-up bounds and other options according to the documentation of solver/minimisation routine
bounds = ((0.045, 0.065),)
op_method = 'L-BFGS-B'
optimisation_options = {'disp': 0,
                        'maxiter': 15000,
                        'maxfun': 15000,
                        'ftol': 2.22e-12,
                        'gtol': 1e-12
                        }
opt_result = minimize(liquid.refinement_objfun, rho_0,
                      bounds=bounds, args=args,
                      options=optimisation_options,
                      method=op_method)
# The solver finds the value of rho that gives the smallest chi^2
rho_refined = opt_result.x[0]

# The interference function can then be re-calculated using the new value for rho,
# and the F(r) refined as above.
interference_func = (liquid.calc_structure_factor(q_data, I_data, composition, rho_refined) -
                     liquid.calc_S_inf(composition, q_data))
args = (q_data, interference_func_0, composition, rho_refined,
        r_min, iter_limit, method, mod_func, window_start, fft_N)
interference_func_2, chi_sq_2 = liquid.calc_impr_interference_func(*args)

# Calculate the corresponding pair distribution functions g(r)
r, g_r_0 = liquid.calc_correlation_func(q_data, interference_func_0, rho_0, dx=dq,
                                        mod_func=mod_func, window_start=window_start,
                                        function='pair_dist_func')
# r values will be the same
_, g_r_1 = liquid.calc_correlation_func(q_data, interference_func_1, rho_0, dx=dq,
                                        mod_func=mod_func, window_start=window_start,
                                        function='pair_dist_func')
_, g_r_2 = liquid.calc_correlation_func(q_data, interference_func_2, rho_refined, dx=dq,
                                        mod_func=mod_func, window_start=window_start,
                                        function='pair_dist_func')

# Plot the data
fig1 = plt.figure()
plt.xlabel(r'Q ($\AA^{-1}$)')
plt.ylabel(r'i(Q)')
# Initial interference function calculation
plt.plot(q_data, interference_func_0, color='g', label=r'Initial i(Q) | $ρ = {rho:.2f}$'.format(rho=rho_0))
# Optimised at rho_0
plt.plot(q_data, interference_func_1, color='r', label=r'Optimised i(Q) | $ρ = {rho:.2f}$ & $χ^{{2}} = {chisq:.2f}$'.format(rho=rho_0, chisq=chi_sq_1))
# Optimised, using refined density estimate
plt.plot(q_data, interference_func_2, color='b', label=r'Optimised i(Q) | $ρ = {rho:.3f}$ & $χ^{{2}} = {chisq:.2f}$'.format(rho=rho_refined, chisq=chi_sq_2))
# Add a legend
plt.legend(loc='best')
# Show the plots
plt.show()

fig2 = plt.figure()
plt.xlabel(r'r ($\AA$)')
plt.ylabel(r'g(r)')
plt.ylim((np.nanmin(g_r_2), np.nanmax(g_r_2)))
window = len(q_data)
# Initial g(r)
plt.plot(r[:window], g_r_0[:window], color='g', label=r'Initial g(r) | $ρ = {rho:.2f}$'.format(rho=rho_0))
# Optimised g(r) at rho_0
plt.plot(r[:window], g_r_1[:window], color='r', label=r'Optimised g(r) | $ρ = {rho:.2f}$ & $χ^{{2}} = {chisq:.2f}$'.format(rho=rho_0, chisq=chi_sq_1))
# Optimised g(r), using refined density estimate
plt.plot(r[:window], g_r_2[:window], color='b', label=r'Optimised g(r) | $ρ = {rho:.3f}$ & $χ^{{2}} = {chisq:.2f}$'.format(rho=rho_refined, chisq=chi_sq_2))
# Add a legend
plt.legend(loc='best')
# Show the plots
plt.show()
